@@ -12,8 +12,117 @@ from .toolsDB import calibration_DB_agent
 from .calibration_object import calibration_object
 from .version import version
 
+from scipy.optimize import leastsq
+import time
 
-class calib_DB:
+
+class calibration_module:
+    """
+    Generic module to apply a calibration on a reader object
+
+    Parameters:
+    -----------
+    reader: km3compass reader object
+      Reader object, either CSK or -something else-
+    moduleID: int
+      DOM module ID inside the file.
+    calibration: ``km3compass.calibration_object``
+      Calibration object to be used.
+    calibrate: bool, default = True
+      Apply or not the calibration. Can be turned off to modify calibration.
+    db_agent: calibration_DB_agent object
+      Provide external calibration_DB_agent object, avoiding additional query to DB
+    """
+
+    def __init__(
+        self,
+        reader,
+        moduleID,
+        calibration,
+        calibrate=True,
+        db_agent=None,
+        verbosity=True,
+    ):
+
+        self.moduleID = moduleID
+        self.calibration = calibration
+        self.db_agent = db_agent
+        self.verbosity = verbosity
+        self.status = False
+        if calibration is not None:
+            self.status = True
+
+        self.df = reader.df
+        self.df = self.df[self.df["DOMID"] == self.moduleID].copy()
+
+        self.mac_address = modID2mac(self.moduleID)
+
+        if self.verbosity:
+            print("DOM mac address : {}".format(self.mac_address))
+
+        if calibrate and calibration is not None:
+            try:
+                self.apply_calibration()
+            except Exception as E:
+                if verbosity:
+                    print(E)
+                self.df = None
+
+    def apply_calibration(self):
+        """
+        Apply calibration on df.
+
+        Equivalent to the implementation in JPP ``JDETECTOR::JCompass::JCompass``
+        https://common.pages.km3net.de/jpp/classJDETECTOR_1_1JCompass.html
+        """
+
+        if self.status == False:
+            raise Exception("Non-valid calibration, can't apply calibration.")
+
+        a = self.df[["AHRS_A0", "AHRS_A1", "AHRS_A2"]].values
+        a_calibrated = (
+            self.calibration.get("A_rot")
+            .dot((a + (-self.calibration.get("A_offsets"))[np.newaxis, :]).T)
+            .T
+        )
+
+        h = self.df[["AHRS_H0", "AHRS_H1", "AHRS_H2"]].values
+        h_calibrated = (
+            self.calibration.get("H_rot")
+            .dot((h + (-self.calibration.get("H_offsets"))[np.newaxis, :]).T)
+            .T
+        )
+
+        for i in range(3):
+            self.df["AHRS_A{}".format(i)] = a_calibrated[:, i]
+            self.df["AHRS_H{}".format(i)] = h_calibrated[:, i]
+
+    def print_calibration(self):
+        """Print a summary of the calibration file"""
+        print(self.calibration)
+
+    def get_summary_df(self):
+        """Return a calibration summary dataframe"""
+        df = pd.DataFrame(
+            {
+                "DOMID": [self.moduleID],
+                "Variant": [self.compassVariant],
+                "CLB UPI": [self.CLB_UPI],
+                "Compass UPI": [self.compass_UPI],
+                "status": [self.status],
+            }
+        ).set_index("DOMID")
+
+        calib = self.row_calib
+
+        if calib is not None:
+            calib["DOMID"] = self.moduleID
+            calib = self.row_calib.to_frame().T.set_index("DOMID")
+
+        return pd.concat((df, calib), axis=1)
+
+
+class calib_DB(calibration_module):
     """
     Module for applying calibration from DB
 
@@ -27,26 +136,21 @@ class calib_DB:
       Apply or not the calibration. Can be turned off to modify calibration.
     db_agent: calibration_DB_agent object
       Provide external calibration_DB_agent object, avoiding additional query to DB
+    verbosity: bool
+      Toggle verbosity during the calibration process
     """
 
     def __init__(self, reader, moduleID, calibrate=True, db_agent=None, verbosity=True):
-        self.moduleID = moduleID
-        self.db_agent = db_agent
-        self.verbosity = verbosity
-        self.df = reader.df
-        self.df = self.df[self.df["DOMID"] == self.moduleID].copy()
 
-        self.mac_address = modID2mac(self.moduleID)
-        if self.verbosity:
-            print("DOM mac address : {}".format(self.mac_address))
+        # Calling parent init without calibration prepare everything
+        # but doesn't apply calibration, ignoring calibrate argument
+        super().__init__(reader, moduleID, None, calibrate, db_agent, verbosity)
 
-        if calibrate:
-            try:
-                self.calibrate()
-            except Exception as E:
-                if verbosity:
-                    print(E)
-                self.df = None
+        # Call DB to load calibration
+        self.load_calibration()
+
+        if calibrate and self.status:
+            self.apply_calibration()
 
     def calibrate(self):
         """Apply calibration chain"""
@@ -80,8 +184,17 @@ class calib_DB:
         try:
             self.compass_UPI = self.db_agent.get_compass_UPI(self.CLB_UPI)
         except:
-            raise KeyError("CLB not found : {}".format(self.CLB_UPI))
-
+            print("Impossible to find compass:")
+            print(f"\tDOM mac address = {self.mac_address}")
+            print(f"\tCLB UPI = {self.CLB_UPI}")
+            self.row_calib = pd.Series(
+                {
+                    "TESTNAME": "Unknown compass",
+                    "DOMID": self.moduleID,
+                    "FIRMWARE_VERSION": "-",
+                }
+            )
+            return
         # Extract compass serial number
         self.compassSN = int(self.compass_UPI.split(".")[-1])
         self.compassVariant = self.compass_UPI.split("/")[1]
@@ -90,70 +203,39 @@ class calib_DB:
         # If yes, it's a base module : no calibration
         if "FLOORID" in self.df.columns:
             if self.df.FLOORID.iloc[0] == 0:
-                raise KeyError(
+                print(
                     "No calibration expected, {} is a base module.".format(
                         self.moduleID
                     )
                 )
+                self.row_calib = pd.Series(
+                    {
+                        "TESTNAME": "No calib (Base module)",
+                        "DOMID": self.moduleID,
+                        "FIRMWARE_VERSION": "-",
+                    }
+                )
+                return
 
         # Get the calib from ahrscalib streamDS
         try:
             self.calibration = self.db_agent.get_calibration(self.compassSN)
             self.row_calib = self.calibration.get("parent")
         except:
-            raise KeyError("No calibration found : {}".format(self.compass_UPI))
+            print("Impossible to find calibration:")
+            print(f"\tDOM mac address = {self.mac_address}")
+            print(f"\tCLB UPI = {self.CLB_UPI}")
+            print(f"\tCompass SN = {self.compassSN}")
+            self.row_calib = pd.Series(
+                {
+                    "TESTNAME": "No calibration",
+                    "DOMID": self.moduleID,
+                    "FIRMWARE_VERSION": "-",
+                }
+            )
+            return
 
-        self.calib = {
-            "magOffset": -self.calibration.get("H_offsets"),
-            "magRot": self.calibration.get("H_rot"),
-            "accRot": self.calibration.get("A_rot"),
-            "accOffset": -self.calibration.get("A_offsets"),
-        }
-
-    def apply_calibration(self):
-        """
-        Apply calibration on df.
-
-        Equivalent to the implementation in JPP ``JDETECTOR::JCompass::JCompass``
-        https://common.pages.km3net.de/jpp/classJDETECTOR_1_1JCompass.html
-        """
-        a = self.df[["AHRS_A0", "AHRS_A1", "AHRS_A2"]].values
-        a_calibrated = (
-            self.calib["accRot"].dot((a + self.calib["accOffset"][np.newaxis, :]).T).T
-        )
-
-        h = self.df[["AHRS_H0", "AHRS_H1", "AHRS_H2"]].values
-        h_calibrated = (
-            self.calib["magRot"].dot((h + self.calib["magOffset"][np.newaxis, :]).T).T
-        )
-
-        for i in range(3):
-            self.df["AHRS_A{}".format(i)] = a_calibrated[:, i]
-            self.df["AHRS_H{}".format(i)] = h_calibrated[:, i]
-
-    def print_calibration(self):
-        """Print a summary of the calibration file"""
-        print(self.calibration)
-
-    def get_summary_df(self):
-        """Return a calibration summary dataframe"""
-        df = pd.DataFrame(
-            {
-                "DOMID": [self.moduleID],
-                "Variant": [self.compassVariant],
-                "CLB UPI": [self.CLB_UPI],
-                "Compass UPI": [self.compass_UPI],
-                "status": [self.row_calib is not None],
-            }
-        ).set_index("DOMID")
-
-        calib = self.row_calib
-
-        if calib is not None:
-            calib["DOMID"] = self.moduleID
-            calib = self.row_calib.to_frame().T.set_index("DOMID")
-
-        return pd.concat((df, calib), axis=1)
+        self.status = True
 
 
 class calib_self_sphere:
@@ -186,6 +268,7 @@ class calib_self_sphere:
 
         self.center = self.fit_result[1:]
         self.radius = self.fit_result[0]
+        self.status = True
         self.apply_calibration()
 
     def sphere_fit(self, spX, spY, spZ):
@@ -294,11 +377,11 @@ class detector_calibration:
       Apply or not the calibration. Can be turned off to modify calibration.
     """
 
-    def __init__(self, reader, detoid, db_agent=None, calib_module=calib_DB):
+    def __init__(self, reader, db_agent=None, calib_module=calib_DB, verbosity=True):
         self.reader = reader
-        self.detoid = detoid
         self.db_agent = db_agent
         self.calib_module = calib_module
+        self.verbosity = verbosity
 
         if self.db_agent is None:
             self.db_agent = calibration_DB_agent()
@@ -318,8 +401,12 @@ class detector_calibration:
         for modID in np.unique(self.reader.df["DOMID"]):
 
             calib = calib_DB(
-                self.reader, modID, db_agent=self.db_agent, verbosity=False
+                self.reader,
+                int(modID),
+                db_agent=self.db_agent,
+                verbosity=self.verbosity,
             )
+
             self.calib_dict[modID] = calib
             # Create a summary row with calibration information
             summary_row = calib.get_summary_df()
@@ -352,7 +439,8 @@ class detector_calibration:
         self.df_overview = self.df_overview.groupby(
             ["DUID", "TESTNAME", "Variant", "FIRMWARE_VERSION"]
         ).sum()["n compass"]
-        print(self.df_overview)
+        print("\n" + "-" * 10 + " Summary per DU " + "-" * 10)
+        print(self.df_overview.to_string())
 
         self.df_overview = self.summary
         self.df_overview.replace({"0.0": "0"}, inplace=True)
@@ -360,7 +448,8 @@ class detector_calibration:
         self.df_overview = self.df_overview.groupby(
             ["TESTNAME", "Variant", "FIRMWARE_VERSION"]
         ).sum()["n compass"]
-        print(self.df_overview)
+        print("\n" + "-" * 10 + " Summary full detector " + "-" * 10)
+        print(self.df_overview.to_string())
 
     def plot_calibration_summary(self):
         """Plot a summary of calibration version per DOM"""
@@ -372,7 +461,6 @@ class detector_calibration:
         fig, axe = plt.subplots()
         axe.set_aspect("equal")
         axe.set_xlim((-1, 19))
-        print(len(DUID_map))
         axe.set_ylim((-1, len(DUID_map)))
 
         nodata_kwargs = {
@@ -423,6 +511,9 @@ class detector_calibration:
             label = "{}, calib V{}, fw {}".format(ind[1], ind[0][-1], ind[2])
             kwargs = {"edgecolor": [0.4] * 3, "facecolor": "C" + str(it), "zorder": 1}
 
+            if ind[2] == "-":
+                label = "{}, no calibration".format(ind[1])
+                kwargs["alpha"] = 0.2
             df = self.summary[
                 (self.summary["TESTNAME"] == ind[0])
                 & (self.summary["Variant"] == ind[1])
@@ -601,3 +692,227 @@ class calibration_ellipsoid_fit:
     def get_calibration(self):
         """Return calibration module"""
         return self.calibration
+
+
+class calib_from_expected_field:
+    """
+    Module that calibrates compasses from sea data
+
+    Parameters:
+    -----------
+    reader: km3compass reader object
+      Reader object, either CSK or -something else-
+    moduleID: int
+      DOM module ID inside the file.
+    """
+
+    def __init__(self, reader, moduleID, compassSN=None):
+        self.df = reader.df
+        du = self.df.loc[self.df["DOMID"] == moduleID, "DUID"].iloc[0]
+        # compass data of other DOMs in same DU for mean calculation
+        du_df = self.df[self.df["DUID"] == du]
+        # a copy that can be accessed in functions
+        self.du_df = du_df
+        # compass data only of DOM we are calibrating
+        self.df = self.df[self.df["DOMID"] == moduleID]
+        self.df = self.df.dropna(subset=["AHRS_H0", "AHRS_H1", "AHRS_H2"])
+
+        # magnetic fields of DOM with ID moduleID
+        h_x = self.df["AHRS_H0"].values
+        h_y = self.df["AHRS_H1"].values
+        h_z = self.df["AHRS_H2"].values
+
+        # select neighbouring DOMs
+        doms = self.select_neigbouring_doms(moduleID)
+        self.db = calibration_DB_agent()
+
+        # here are computed the mean magnetic fields in each direction and the field norm
+        h_x_mean, h_y_mean, h_z_mean, r_mean = self.mean_mag_fields(
+            int(moduleID), doms, reader
+        )
+
+        # here we subtract this mean field in each direction
+        mean_x = float(np.mean(h_x)) - h_x_mean
+        mean_y = float(np.mean(h_y)) - h_y_mean
+        mean_z = float(np.mean(h_z)) - h_z_mean
+
+        # least square fitting done here
+        """
+        NB doing the lsq fit actually worsens the results
+           commented out for now
+        self.fit_result = self.sphere_fit(
+            h_x,
+            h_y,
+            h_z,
+            r_mean, mean_x, mean_y, mean_z
+        )
+        """
+        fit_result = np.array([r_mean, mean_x, mean_y, mean_z])
+        self.center = fit_result[1:]
+        self.radius = fit_result[0]
+        # storing in calibration object
+        self.compass_SN = compassSN
+
+        # If no compass SN provided, get it from moduleID
+        if self.compass_SN is None:
+            self.compass_UPI = self.db.get_compass_UPI_from_modID(moduleID)
+            self.compass_SN = int(self.compass_UPI.split(".")[-1])
+
+        self.make_calibration_object(moduleID, reader)
+
+    def select_neigbouring_doms(self, moduleID, n_neighbors=2):
+        """
+        select neigbouring DOMs to the one currently calibrated
+        Parameters:
+        -----------
+        moduleID: int
+          DOM module ID inside the file.
+        n_neighbors: int (defaul: 2)
+          Number of DOMs taken on each side.
+        Return:
+        -------
+        doms: numpy array
+          array of neigbouring DOMIDs
+        """
+        du_df = self.du_df
+        floor = int(du_df.loc[du_df["DOMID"] == moduleID, "FLOORID"].iloc[0])
+
+        if floor < 1 + n_neighbors:
+            floor = 1 + n_neighbors
+
+        elif floor > 18 - n_neighbors:
+            floor = 18 - n_neighbors
+
+        du_df = du_df[
+            (du_df["FLOORID"] <= floor + n_neighbors)
+            & (du_df["FLOORID"] >= floor - n_neighbors)
+        ]
+
+        doms = du_df.DOMID.unique().astype(int)
+        return doms
+
+    def mean_mag_fields(self, moduleID, doms, reader, cut_radius=0.015):
+        """
+        compute mean magnetic field in each direction, using calibrated magnetometer data
+        uses only compasses in the same DU (passed with doms array, contains dom IDs in same DU)
+        uses only compasses of the same variant
+        rejects miscalibrated DOMs: the norm of the magnetic field is used as a control parameter
+        for miscalibrated DOMs
+        Parameters:
+        -----------
+        moduleID: int
+          DOM module ID inside the file.
+        doms: numpy array
+          The neigbouring DOMs to use in computation of the expected field
+        reader: km3compass reader object
+          Reader object, either CSK or -something else-
+        Return:
+        -------
+        mean H_0, mean H_1, mean H_2, mean magnetic field norm
+         The mean magnetic field along each axis and the mean norm as experienced by neighbouring DOMs
+        """
+        moduleID = int(moduleID)
+        # Calibrate all the DOMs available in the reader
+        calib_df = detector_calibration(reader, verbosity=False)
+        calib_df.apply_calibration()
+
+        df = calib_df.df
+        df["DOMID"] = df["DOMID"].astype(int)
+
+        # field norms
+        df["AHRS_HR"] = np.sqrt(
+            df["AHRS_H0"] ** 2 + df["AHRS_H1"] ** 2 + df["AHRS_H2"] ** 2
+        )
+
+        doms = [
+            d for d in doms if d != moduleID
+        ]  # Generate the list of DOMs, excluding target DOM
+        # check if the neigbouring DOMs are correctly calibrated
+        for dom in doms:
+            to_drop = (
+                calib_df.summary.loc[dom]["status"] == False
+                or df[df["DOMID"] == dom]["AHRS_HR"].std() >= cut_radius
+            )
+            print(dom, to_drop)
+            if to_drop:
+                df.drop(df[df["DOMID"] == dom].index)
+
+        # Reduce the df to one line per DOM, that contains the mean value of the field
+        df = df.groupby("DOMID").mean().reset_index()
+        # Use the CLBUPI, added when calibrating, to add the compassUPI
+        def fail_proof_compass_UPI(domID):
+            try:
+                return self.db.get_compass_UPI_from_modID(domID)
+            except Exception as E:
+                print(f"Impossible to get Compass UPI for {domID}: {E}")
+                return np.nan
+
+        df["Compass_UPI"] = df["DOMID"].apply(fail_proof_compass_UPI)
+        df = df.dropna(subset=["Compass_UPI"])
+        # Declare and use a function to convert compass_UPI to
+        def CompassUPI_to_variant(upi):
+            return upi.split("/")[1]
+
+        df["Compass_variant"] = df["Compass_UPI"].apply(CompassUPI_to_variant)
+
+        df = df.set_index("DOMID")
+
+        temp_array = df.index.to_numpy(dtype=int)
+        # check if all elements of array doms are in dataframe
+        for dom in list(doms):
+            ind = np.where(temp_array == dom)
+            ind = np.array(ind)
+            if ind.size == 0:
+                ind2 = np.where(doms == dom)
+                doms = np.delete(doms, ind2)
+        doms = [float(i) for i in doms]
+        target_dom = df.loc[moduleID]
+        df = df.loc[doms]  # Reduce the list to the neighbor DOMs
+        return (
+            df["AHRS_H0"].mean(),
+            df["AHRS_H1"].mean(),
+            df["AHRS_H2"].mean(),
+            df["AHRS_HR"].mean(),
+        )
+
+    def make_calibration_object(self, moduleID, reader):
+        """
+        creates calibration object
+
+        Parameters:
+        -----------
+        moduleID: int
+          DOM module ID inside the file.
+        reader: km3compass reader object
+          Reader object, either CSK or -something else-
+        """
+
+        calib_df = calib_DB(reader, moduleID, calibrate=True)
+        """
+        get temporary calibration object to fetch calibration constants we do not compute:
+        A_norm, A_offsets, A_rot
+        H_norm, H_rot: if compass fails acceptance tests, can we trust these parameters?
+        """
+        self.calibration = None
+        calibration_temp = self.db.get_calibration(self.compass_SN)
+        self.A_norm = calibration_temp.get("A_norm")
+        self.A_offsets = calibration_temp.get("A_offsets")
+        self.A_rot = calibration_temp.get("A_rot")
+        # self.H_norm = calibration_temp.get("H_norm")
+        self.H_rot = calibration_temp.get("H_rot")
+        self.H_offsets = self.center
+        self.H_norm = self.radius
+
+        # NB: change calibration type to some agreed-upon value
+        self.calibration = calibration_object(
+            compass_SN=self.compass_SN,
+            compass_UPI=self.compass_UPI,
+            type="AHRS-SEA-CALIBRATION-v1",
+            source=f"km3compass-{version}",
+            A_norm=self.A_norm,
+            A_offsets=-self.A_offsets,
+            A_rot=self.A_rot,
+            H_norm=1.0,
+            H_offsets=-self.H_offsets,
+            H_rot=np.identity(3),
+        )
