@@ -21,7 +21,7 @@ class readerCSK:
         self.filename = filename
         self.rawframe = load_CSK(self.filename)
         self.df = pd.DataFrame(self.rawframe)
-        self.df["datetime"] = pd.to_datetime(self.df["time"], unit="s")
+        self.df["datetime"] = pd.to_datetime(self.df["time"], unit="s").fillna(pd.NaT)
         # Print some generic informations about the file
         print("File loaded, {} rows".format(self.df.shape[0]))
         self.module_IDs = np.unique(self.df["DOMID"])
